# Copyright (C) 2010 W. Trevor King <wking@drexel.edu>
#
# This file is part of ChemDB.
#
# ChemDB is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# ChemDB is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ChemDB.  If not, see <http://www.gnu.org/licenses/>.

"""Chemical inventory web interface.
"""

from __future__ import with_statement

from cgi import escape
import os
import os.path
import re
import sys, logging

import cherrypy
from jinja2 import Environment, FileSystemLoader

from .daemon import Daemon
from .db.text import TextDB
from .chemdb import valid_CASno, MSDSManager, DocGen


__version__ = "0.4"


class Server (object):
    """ChemDB web interface."""
    def __init__(self, db=None, MSDS_manager=None, docs=None,
                 template_root='template', static_dir='static',
                 static_url='static'):
        self.db = db
        self.MSDS_manager = MSDS_manager
        self.docs = docs
        self.static_dir = static_dir
        self.static_url = static_url
        self.env = Environment(loader=FileSystemLoader(template_root))
        self.width_max = {'CAS#':12}
        self.entry_width = 50  # display width of entry fields
        self.raw_fields = ['ID','Name']
        self.MSDS_content_types = [
            'text/plain', 'text/html', 'application/pdf']

    @cherrypy.expose
    def index(self, id=None):
        """Chemical index page.
        """
        self.db._refresh()
        if id == None:
            return self._index_page()
        else:
            return self._view_page(id=id)

    def _index_page(self):
        records = self.db.records()
        for record in records:
            for field in self.db.field_list():
                rstring = record[field]
                if field in self.width_max: # truncate if width is regulated...
                    w = self.width_max[field]
                    if len(rstring) > w:    # ... and full string is too long.
                        rstring = "%s..." % rstring[:w-3]
                # if the field isn't raw, protect special chars
                if not field in self.raw_fields:
                    rstring = escape(rstring)
            if self.MSDS_manager.has_MSDS(record['db_id']):
                # link the id to the MSDS
                record['Name'] = (
                    '<a href="%s">%s</a>'
                    % (self._MSDS_path(record['db_id']), record['Name']))
        template = self.env.get_template('index.html')
        return template.render(fields=self.db.field_list(), records=records)

    def _view_page(self, id):
        db_id = int(id)
        record = self.db.record(db_id)
        MSDS_link = None
        if self.MSDS_manager.has_MSDS(db_id):
            path = self._MSDS_path(db_id)
            MSDS_link = '<a href="%s">%s</a>' % (
                path, os.path.basename(path))
        template = self.env.get_template('record.html')
        return template.render(fields=self.db.field_list(),
                               long_fields=self.db.long_fields(),
                               record=record,
                               MSDS=MSDS_link,
                               escape=escape)

    def _MSDS_path(self, db_id):
        path = self.MSDS_manager.get_MSDS_path(db_id)
        path = os.path.relpath(path, self.static_dir)
        path = os.path.join(self.static_url, path)
        return path

    @cherrypy.expose
    def edit(self, id=None, **kwargs):
        """Render the edit-record.html template for the specified record.
        """
        self.db._refresh()
        if kwargs:
            self._update_record(id=id, **kwargs)
            raise cherrypy.HTTPRedirect(u'.?id=%s' % id, status=303)
        MSDSs = self.MSDS_manager.get_all(simlinks=False)
        if id in [None, '-1']:
            record = dict([(field,'') for field in self.db.field_list()])
            record['db_id'] = '-1'
            record['ID'] = str(record['db_id'])
        else:
            db_id = int(id)
            record = self.db.record(db_id)
            if self.MSDS_manager.has_MSDS(db_id):
                MSDSs = None  # only ask for an MSDS if we still need one
        template = self.env.get_template('edit-record.html')
        return template.render(fields=self.db.field_list(),
                               long_fields=self.db.long_fields(),
                               record=record,
                               MSDSs=MSDSs,
                               entry_width=self.entry_width,
                               escape=escape)

    def _update_record(self, id=None, **kwargs):
        """Update a record with form input.
        """
        self.db._refresh()
        if id in [None, '-1']:
            record = self.db.new_record()
            logging.info('new record %s' % record['db_id'])
        else:
            db_id = int(id)
            record = self.db.record(db_id)
        update = False
        for field in self.db.field_list():
            if kwargs[field] != record[field]:
                # TODO: add validation!
                update = True
                record[field] = kwargs[field]
        if kwargs.get('MSDS source', None) == 'upload':
            # Handle any MSDS file actions
            f = kwargs['MSDS upload']
            contents = f.file.read()
            if len(contents) > 0 and f.type in self.MSDS_content_types:
                self.MSDS_manager.save(int(record['ID']), contents, f.type)
        elif kwargs.get('MSDS source', None) == 'share':
            logging.info('linking MSDS %d to %d'
                         % (int(record['ID']),
                            int(kwargs['MSDS share'])) )
            self.MSDS_manager.link(int(record['ID']),
                                   int(kwargs['MSDS share']))
        if update:
            self.db.set_record(record['db_id'], record, backup=True)


class Docs (object):
    "Generate and serve assorted official documents."
    def __init__(self, db=None, docgen=None, template_root='template'):
        self.db = db
        self.docgen = docgen
        self.namewidth = 40
        self.env = Environment(loader=FileSystemLoader(template_root))

    @cherrypy.expose
    def index(self):
        """List the available documents.
        """
        template = self.env.get_template('docs.html')
        return template.render()

    @cherrypy.expose
    def inventory_pdf(self):
        self.db._refresh()
        path = self.docgen.inventory(namewidth=self.namewidth)
        cherrypy.response.headers['Content-Type'] = 'application/pdf'
        return file(path, 'rb').read()

    @cherrypy.expose
    def door_warning_pdf(self, location=None):
        self.db._refresh()
        if location == None:
            valid = lambda r: r['Disposed'] == ''
        else:
            regexp = re.compile(location, re.I) # Case insensitive
            valid = lambda r: r['Disposed'] == '' and regexp.match(r['Location'])
        path = self.docgen.door_warning(valid_record=valid)
        cherrypy.response.headers['Content-Type'] = 'application/pdf'
        return file(path, 'rb').read()


class ServerDaemon (Daemon):
    def __init__(self):
        super(ServerDaemon, self).__init__()
        self.gid = None
        self.uid = None
        self.pidfile = './chem_web.pid'
        self.logfile = './chem_web.log'
        self.loglevel = 'INFO'

    def run(self):
        if cherrypy.__version__.startswith('3.'):
            cherrypy.quickstart(root=self.server, config=self.app_config)

#        if self.pkey_file == None or self.cert_file == None:
#            logging.info("http://%s:%d/" % web.validip(self.ip_address))
#        else:
#            logging.info("https://%s:%d/" % web.validip(self.ip_address))
#
#        webpy_func = web.webpyfunc(urls, globals(), False)
#        wsgi_func = web.wsgifunc(webpy_func)
#        web.httpserver.runsimple(wsgi_func,
#                                 web.validip(self.ip_address),
#                                 ssl_private_key_filename=self.pkey_file,
#                                 ssl_certificate_filename=self.cert_file)

    def read_basic_config(self):
        pass

    def parse_options(self):
        from optparse import OptionParser
            
        usage_string = ('%prog [options]\n'
                        '\n'
                        '2008, W. Trevor King.\n')
        version_string = '%%prog %s' % __version__
        p = OptionParser(usage=usage_string, version=version_string)

        # Non-server options
        p.add_option('-t', '--test', dest='test', default=False,
                     action='store_true', help='Run internal tests and exit.')

        # Daemon options
        p.add_option('--start', dest='action',
                     action='store_const', const='start', default='start',
                     help='Start the daemon (the default action).')
        p.add_option('--stop', dest='action',
                     action='store_const', const='stop', default='start',
                     help='Stop the daemon.')
        p.add_option('-n', '--nodaemon', dest='daemonize',
                     action='store_false', default=True,
                     help='Run in the foreground.')

        # Server options
        p.add_option('-a', '--address', dest='address', default='127.0.0.1',
                     metavar='ADDR',
                     help='Address that the server will bind to (%default).')
        p.add_option('-p', '--port', dest='port', default=8080, type='int',
                     metavar='PORT',
                     help='Port that the server will listen on (%default).')
        p.add_option('-s', '--secure', dest="secure",
                     help="Run in secure (HTTPS) mode.",
                     type='string', metavar="PKEY_FILE:CERT_FILE")
        p.add_option('-v', '--verbose', dest="verbose", action="store_true",
                     help="Print lots of debugging information.",
                     default=False)
        p.add_option('--static', dest='static', metavar='PATH',
                     help="Path to the directory of static files (%default).",
                     default=os.path.join('example', 'static'))
        p.add_option('--template', dest='template', metavar='PATH',
                     help="Path to the directory of template files (%default).",
                     default=os.path.join('template', 'web'))
        p.add_option('--doc', dest='doc', metavar='PATH',
                     help="Path to the directory of document generation files (%default).",
                     default=os.path.join('template', 'doc'))
        p.add_option('--htaccess', dest='htaccess', metavar='FILE',
                     help="Path to the htaccess file (%default).",
                     default='.htaccess')

        # Database options
        p.add_option('--database', dest='database', metavar='FILE',
                     help="Path to the database file (%default).",
                     default=os.path.join('example', 'inventory.db'))

        self.options, args = p.parse_args()
        p.destroy()

        if self.options.test == True:
            _test()
            sys.exit(0)

        if self.options.verbose:
            self.loglevel = 'DEBUG'

        # get self.options for httpserver
        if self.options.secure != None:
            split = self.options.secure.split(':', 1)
            assert len(split) == 2, (
                "Invalid secure argument '%s'" % self.options.secure)
            self.pkey_file = split[0]
            self.cert_file = split[1]

        # HACK! to ensure we *always* get utf-8 output
        #reload(sys)
        #sys.setdefaultencoding('utf-8')

        dirname,filename = os.path.split(
            os.path.abspath(self.options.database))
        static_dir = os.path.abspath(self.options.static)
        MSDS_dir = os.path.join(static_dir, 'MSDS')
        template_dir = os.path.abspath(self.options.template)
        doc_dir = os.path.abspath(self.options.doc)        
        db = TextDB(filename=filename, current_dir=dirname)
        MSDS_manager = MSDSManager(db=db, dir=MSDS_dir)
        docgen = DocGen(db=db, doc_root=doc_dir)
        docs = Docs(db=db, docgen=docgen, template_root=template_dir)
        server = Server(db=db, MSDS_manager=MSDS_manager, docs=docs,
                        template_root=template_dir, static_dir=static_dir,
                        static_url='/static')

        if cherrypy.__version__.startswith('3.'):
            cherrypy.config.update({ # http://www.cherrypy.org/wiki/ConfigAPI
                    'server.socket_host': self.options.address,
                    'server.socket_port': self.options.port,
                    'tools.decode.on': True,
                    'tools.encode.on': True,
                    'tools.encode.encoding': 'utf8',
                    'tools.staticdir.root': static_dir,
                    })
            digest_auth = None
            if cherrypy.__version__.startswith('3.2'):
                try:
                    get_ha1 = cherrypy.lib.auth_digest.get_ha1_file_htdigest(
                        self.options.htaccess)
                except IOError:
                    pass
                else:
                    digest_auth = {
                        'tools.auth_digest.on': True,
                        'tools.auth_digest.realm': 'chemdb',
                        'tools.auth_digest.get_ha1': get_ha1,
                        'tools.auth_digest.key': str(uuid.uuid4()),
                        }
            else:
                passwds = {}
                try:
                    with open(self.options.htaccess, 'r') as f:
                        for line in f:
                            user,realm,ha1 = line.strip().split(':')
                            passwds[user] = ha1  # use the ha1 as the password
                except IOError:
                    pass
                else:
                    digest_auth = {
                        'tools.digest_auth.on': True,
                        'tools.digest_auth.realm': 'chemdb',
                        'tools.digest_auth.users': passwds,
                        }
            app_config = {
                '/static': {
                    'tools.staticdir.on': True,
                    'tools.staticdir.dir': '',
                    },
                }
            if digest_auth != None:
                for url in ['edit']:
                    app_config[url] = digest_auth
            self.server = server
            self.app_config = app_config
        else:
            raise NotImplementedError(
                'Unsupported CherryPy version %s' % cherrypy.__version__)


def _test():
    import doctest
    doctest.testmod()
